"""Operators"""

from __future__ import annotations

# from itertools import islice
from typing import TYPE_CHECKING

from ..sets.cases import Elem, FCase
from ..sets.function import F

if TYPE_CHECKING:
    from ..sets.index import I
    from ..sets.variable import V


def sigma(variable: V, over: I = None, position: int = None) -> F:
    """
    Summation without recursion, also allows better printing

    :param variable: variable set
    :type variable: V
    :param over: over what index
    :type over: I | None
    :param position: position of the index in the variable indices. Defaults to None.
    :type position: int | None

    :returns: summed up function
    :rtype: F
    """

    def _determine_index(over, position):
        if over:
            length = len(over)

            if not position:
                position = variable.index.index(over)

            # Precompute slices
            before = variable.index[:position]
            after = variable.index[position + 1 :]

            # Build variables
            _variables = [
                variable(*before, _index, *after, make_new=True) for _index in over
            ]

        else:
            # sum over the entire set
            _variables = variable._
            length = len(variable)
            position = None
            over = variable.index

        return _variables, over, length, position

    def _birth_parent(variable, _variables, over, position, length):
        _f = F()
        _f.case = FCase.SUM
        _f.issumhow = (variable.copy(), over, position)
        _f.variables = _variables
        _f.index = tuple(v.index for v in _f.variables)
        _f.one = _f
        _f.one_type = Elem.F
        _f.give_name()

        _f.rhs_thetas = []
        length_var = len(_variables[0])
        keys = list(zip(*(v.map for v in _f.variables)))
        _f.A = [[1] * length for _ in range(length_var)]
        return _f, length_var, keys

    def _birth_children(f, length_var, keys):
        for n in range(length_var):
            # make the child functions
            f_child = F()
            f_child.variables = [v[n] for v in f.variables]
            f_child.P = [v.n for v in f_child.variables]
            f_child.A = [1] * length
            key = keys[n]
            f_child.issumhow = (variable[length * n], over, position)
            f_child.parent = f
            f_child.case = FCase.SUM
            f_child.rhs_thetas = []
            f.P.append(f_child.P)
            f_child.give_name()
            f_child._ = [f_child]
            f._.append(f_child)
            f.map[key] = f_child
            f_child.map[key] = f_child
            f_child.parent = f
            f_child.index = key
            f_child.one = f_child
            f_child.one_type = Elem.F
        return f

    _variables, over, length, position = _determine_index(over, position)

    if length == 2:
        # short-circuit for length 2
        # just v_0 + v_1
        return _variables[0] + _variables[1]

    f, length_var, keys = _birth_parent(variable, _variables, over, position, length)

    return _birth_children(f, length_var, keys)



"""Plotting Utilities"""

from __future__ import annotations

from typing import TYPE_CHECKING

from matplotlib import pyplot as plt
from matplotlib import rc

if TYPE_CHECKING:
    from ..sets.parameter import P
    from ..sets.variable import V


def draw(
    element: V | P,
    data: list[float] | None = None,
    kind: str = "line",
    font_size: float = 16,
    fig_size: tuple[float, float] = (12, 6),
    linewidth: float = 0.7,
    color: str = "blue",
    grid_alpha: float = 0.3,
    usetex: bool = True,
    str_idx_lim: int = 10,
):
    """
    Plot the variable set

    :param kind: Type of plot ['line', 'bar']. Defaults to 'line'.
    :type kind: str, optional
    :param font_size: Font size for the plot. Defaults to 16.
    :type font_size: float, optional
    :param fig_size: Size of the figure. Defaults to (12, 6).
    :type fig_size: tuple[float, float], optional
    :param linewidth: Width of the line in the plot. Defaults to 0.7.
    :type linewidth: float, optional
    :param color: Color of the line in the plot. Defaults to 'blue'.
    :type color: str, optional
    :param grid_alpha: Transparency of the grid lines. Defaults to 0.3.
    :type grid_alpha: float, optional
    :param usetex: Use LaTeX for text rendering. Defaults to True.
    :type usetex: bool, optional
    :param str_idx_lim: Limit for string indices display. Defaults to 10.
    :type str_idx_lim: int, optional
    """

    ax = plt.subplots(figsize=fig_size)[1]

    # the values are the y-axis
    y = data

    _len = len(y)

    # the indices are the x-axis
    if _len <= str_idx_lim:
        x = [str(idx) for idx in element.map]
    else:
        x = list(range(len(y)))

    if usetex:
        rc(
            "font",
            **{"family": "serif", "serif": ["Computer Modern"], "size": font_size},
        )
        rc("text", usetex=usetex)
    else:
        rc("font", **{"size": font_size})

    if kind == "line":
        ax.plot(x, y, linewidth=linewidth, color=color)

    elif kind == "bar":
        ax.bar(x, y, linewidth=linewidth, color=color)

    ax.set_title(
        rf"${element.latex()}$",
    )
    ax.set_ylabel(r"Values")
    ax.set_xlabel(r"Indices")
    ax.grid(alpha=grid_alpha)

    if _len <= str_idx_lim:
        ax.set_xticks(x)
        ax.set_xticklabels(
            [
                rf"${tuple([idx.ltx for idx in index])}$".replace("'", "").replace(
                    "\\", ""
                )
                for index in element.map
            ]
        )

    plt.rcdefaults()
    return plt



"""Decorators for functions"""

import logging
import time
from functools import wraps


def once(func):
    """Ensures the function is executed only once"""

    def wrapper(*args, **kwargs):
        if not hasattr(wrapper, "has_run"):
            wrapper.has_run = True
            return func(*args, **kwargs)

    return wrapper


def timer(
    logger: logging.Logger, kind: str = None, with_return=True, level=logging.INFO
):
    """
    Logs execution time and optionally shows a full computation using function arguments and result.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            start = time.time()
            # returns the result if successful, else False
            result = func(*args, **kwargs)
            elapsed = time.time() - start
            if result:
                if kind == 'generate-mps':
                    msg = f"📝  Generated {result}.mps"

                if kind == 'generate-solution':
                    msg = f"📝  Generated Solution object for {result}. See .solution"

                if kind == 'generate-ppopt':
                    msg = "📝  Generated MPLP. See .formulation"

                if kind == 'generate-gurobi':
                    msg = "📝  Generated gurobipy model. See .formulation"

                if kind == 'solve-mpqp':
                    msg = "✅  Solved MPLP using PPOPT. See .solution"

                if kind == 'optimize':
                    msg = f"✅  {result[0]} optimized using {result[1]}. Display using .output()"

                logger.log(
                    level,
                    f"{msg:<75} ⏱ {elapsed:.4f} s",
                )

            if with_return:
                return result

        return wrapper

    return decorator
